/*
 * AugmentedConstituent.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 25 Apr 2002
 * Time-stamp: <Fri May 23 12:17:37 EDT 2003 ferguson>
 */

/**
 * An AugmentedConstituent of a parse represents a syntactic element
 * augmented with feature information.
 * It consists of:
 * <UL>
 * <LI>the syntactic category
 * <LI>a set of features
 * </UL>
 */
public class AugmentedConstituent extends Constituent implements Unifiable {
    /**
     * Set of features of this AugmentedConstituent.
     */
    FeatureSet features;
    /**
     * Create a new AugmentedConstituent with the given attributes.
     */
    public AugmentedConstituent(Symbol category, FeatureSet feats) {
	super(category);
	this.features = feats;
    }
    /**
     * Returns the FeatureSet of this AugmentedConstituent.
     */
    public FeatureSet getFeatures() {
	return features;
    }
    /**
     * Returns the printable representation of this AugmentedConstituent as a
     * String.
     */
    public String toString() {
	return "(" + category.toString() + (features.isEmpty() ? "" : (" " + features.toString())) + ")";
    }
    /**
     * Returns true if this AugmentedConstituent unifies with the given
     * one, and puts the unifying substitution in bindings if so.
     * Constituents can only unify with other constituents, which they
     * do if their category and features both unify.
     */
    public boolean unifyWith(Unifiable other, BindingSet bindings) {
	Debug.debug("AugmentedConstituent.unifyWith:\n  this=" + this + "\n  other=" + other);
	if (!(other instanceof AugmentedConstituent)) {
	    return false;
	} else {
	    AugmentedConstituent augother = (AugmentedConstituent)other;
	    return (category.equals(augother.category) &&
		    features.unifyWith(augother.features, bindings));
	}
    }
    /**
     * Returns a new AugmentedConstituent which is a copy of this one.
     */
    public Unifiable copyWithBindings(BindingSet bindings) {
	Debug.debug("AugmentedConstituent.copy: " + this);
	return new AugmentedConstituent(category,
					features.copyWithBindings(bindings));
    }
}
