/*
 * FeatureSet.java
 *
 * George Ferguson, ferguson@cs.rochester.edu,  1 May 2002
 * Time-stamp: <Wed May 14 12:38:13 EDT 2003 ferguson>
 */

import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;

/**
 * A FeatureSet is a set of name/value pairs representing the
 * features of a constituent. The names are Symbols. The values
 * are Unifiables (Symbol, Variable, or Expression).
 * <P>
 * Implementation note: The traditional implementation of this
 * is an alist, so we use an AssociationList. Since feature set
 * unification is the most important thing we do, it might make
 * sense to pick a better data structure. For example, since we
 * call entrySet() on each call to unifyWith(), we should probably
 * avoid the cost of allocating those items over and over and just
 * keep the names and values in parallel lists (which would be
 * linear time to lookup but no extra cost to iterate over, and anyway
 * the feature sets are small).
 */
public class FeatureSet extends AssociationList {
    /**
     * Set the value of the given feature in this FeatureSet.
     */
    public void setFeatureValue(Symbol name, Unifiable value) {
	put(name, value);
    }
    /**
     * returns the value of the given feature in this FeatureSet.
     */
    public Unifiable getFeatureValue(Symbol name) {
	return (Unifiable)get(name);
    }
    /**
     * Returns the String representation of this FeatureSet.
     */
    public String toString() {
	StringBuffer buf = new StringBuffer();
	Iterator entries = entrySet().iterator();
	try {
	    buf.append("(");
	    while (entries.hasNext()) {
		Map.Entry entry = (Map.Entry)entries.next();
		buf.append("(");
		buf.append(entry.getKey());
		buf.append(" ");
		buf.append(entry.getValue());
		buf.append(")");
		if (entries.hasNext()) {
		    buf.append(" ");
		}
	    }
	    buf.append(")");
	} catch (NoSuchElementException ex) {
	}
	return buf.toString();
    }
    /**
     * Returns true if this FeatureSet unifies with the given one,
     * and puts the unifying substitution in the given BindingSet.
     * Two FeatureSets unify if the values of features present in
     * both sets unify (features in only one set don't need to be
     * checked).
     */
    public boolean unifyWith(FeatureSet other, BindingSet bindings) {
	//Debug.debug("unifyWith: " + this + " with " + other);
	// For each of my features...
	Iterator entries = entrySet().iterator();
	try {
	    while (entries.hasNext()) {
		Map.Entry entry = (Map.Entry)entries.next();
		Symbol name = (Symbol)entry.getKey();
		// ...Get my value for the feature
		Unifiable value = (Unifiable)entry.getValue();
		// And get the other value for the feature
		Unifiable otherValue = other.getFeatureValue(name);
		// If it's specified but they don't unify...
		if (otherValue != null &&
		    !value.unifyWith(otherValue, bindings)) {
		    // ...Then we fail, otherwise carry on
		    //Debug.debug("unifyWith: failed");
		    return false;
		}
	    }
	} catch (NoSuchElementException ex) {
	}
	// Got here => all my features unified with other's
	//Debug.debug("unifyWith: successful: " + bindings);
	return true;
    }
    /**
     * Returns a copy of this FeatureSet with the given bindings applied.
     */
    public FeatureSet copyWithBindings(BindingSet bindings) {
	Debug.debug("FeatureSet.copy: " + this);
	FeatureSet result = new FeatureSet();
	Iterator entries = entrySet().iterator();
	try {
	    while (entries.hasNext()) {
		Map.Entry entry = (Map.Entry)entries.next();
		Symbol name = (Symbol)entry.getKey();
		Unifiable value = (Unifiable)entry.getValue();
		Debug.debug("FeatureSet.copy: copying feature " + name);
		result.setFeatureValue(name, value.copyWithBindings(bindings));
	    }
	} catch (NoSuchElementException ex) {
	}
	Debug.debug("FeatureSet.copy: result is " + result);
	return result;
    }
}
