/*
 * BottomUpChartParser.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 13 May 2002
 * Time-stamp: <Fri May 23 11:31:51 EDT 2003 ferguson>
 */

import java.io.BufferedReader;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * This is the simple bottom-up parser described in NLU Chapter 3.
 * It uses just syntactic catgeories (no features).
 */
public class BottomUpChartParser {
    /**
     * The Lexicon used by this BottomUpChartParser.
     */
    Lexicon lexicon;
    /**
     * The Grammar used by this BottomUpChartParser.
     */
    Grammar grammar;
    /**
     * Set the Lexicon used by this BottomUpChartParser.
     */
    void setLexicon(Lexicon lexicon) {
	this.lexicon = lexicon;
    }
    /**
     * Set the Grammar used by this BottomUpChartParser.
     */
    void setGrammar(Grammar grammar) {
	this.grammar = grammar;
    }
    /**
     * The Tokenizer used by this BottomUpChartParser.
     */
    Tokenizer tokenizer;
    /**
     * Set the Tokenizer used by this BottomUpChartParser.
     */
    void setTokenizer(Tokenizer tokenizer) {
	this.tokenizer = tokenizer;
    }
    /**
     * Number of parses after which we should stop (0 means find all parses).
     */
    int nParsesWanted = 0;
    /**
     * Set the number of parses after which to stop (0 means find all parses).
     */
    void setNParsesWanted(int n) {
	nParsesWanted = n;
    }
    /**
     * Create a new BottomUpChartParser.
     */
    public BottomUpChartParser() {
	setLexicon(new Lexicon());
	setGrammar(new Grammar());
	setTokenizer(new Tokenizer());
    }
    /**
     * The Chart used by this BottomUpChartParser.
     */
    Chart chart;
    /**
     * The Agenda used by this BottomUpChartParser.
     */
    Agenda agenda;
    /**
     * The input tokens being parsed (for debugging).
     */
    List tokens;
    /**
     * Length of the sentence being parsed.
     */
    int inputLength;
    /**
     * Number of parses found so far.
     */
    int nParsesFound;
    /**
     * Parse the given input string.
     * This is the main interface method for the BottomUpChartParser.
     * After parsing, the chart is left unchanged until the next call
     * to this method.
     * Returns the number of parses found.
     */
    public int parse(String input) {
	tokens = tokenizer.tokenizeInput(input);
	inputLength = tokens.size();
	chart = new Chart(inputLength);
	agenda = new Agenda();
	initializeAgenda(tokens);
	nParsesFound = 0;
	while (!parseDone()) {
	    addEntryToChart(agenda.getNext());
	}
	return nParsesFound;
    }
    /**
     * Adds initial entries to the agenda for the given (tokenized) input
     * sentence.
     */
    void initializeAgenda(List tokenList) {
	Iterator tokens = tokenList.iterator();
	try {
	    int pos = 0;
	    while (tokens.hasNext()) {
		Symbol word = (Symbol)tokens.next();
		initializeAgendaForWord(word, pos);
		pos += 1;
	    }
	} catch (NoSuchElementException ex) {
	}
    }
    /**
     * Finds the entry in the lexicon for a specific word and creates
     * chart entries on the agenda for each interpretation.
     */
    void initializeAgendaForWord(Symbol word, int pos) {
	List entries = lexicon.getEntriesForWord(word);
	if (entries != null && entries.size() > 0) {
	    Iterator e = entries.iterator();
	    try {
		while (e.hasNext()) {
		    Constituent constit = (Constituent)e.next();
		    Symbol category = constit.getCategory();
		    Symbol id = createEntryId(category);
		    Entry entry = new Entry(id, constit, pos, pos + 1);
		    // Add to front so words get handled in input order
		    agenda.add(0, entry);
		}
	    } catch (NoSuchElementException ex) {
	    }
	} else {
	    Debug.warn("unknown word: " + word + " at position " + pos);
	}
    }
    /**
     * Creates a Symbol id for a new entry of the given category.
     * This could just call Symbol.gensym(category), but with a bit
     * more bookkeeping we can make the numbering come out as it
     * does in the NLU textbook (essentially, per-category numbering).
     */
    Symbol createEntryId(Symbol category) {
	Integer count = (Integer)categoryCounters.get(category);
	if (count == null) {
	    count = new Integer(0);
	}
	count = new Integer(count.intValue() + 1);
	categoryCounters.put(category, count);
	return Symbol.intern(category.toString() + count.toString());
    }
    /**
     * Set of counters to support more intuitive use of gensym().
     */
    HashMap categoryCounters = new HashMap();
    /**
     * Returns true if we should stop parsing, either because
     * the agenda is empty, or because we have found the requested
     * number of parses.
     */
    boolean parseDone() {
	return (agenda.isEmpty() ||
		(nParsesWanted > 0 && nParsesFound >= nParsesWanted));
    }
    /**
     * Adds the given Entry (completed constituent) to the chart and
     * tries to add or extend arcs (possibly adding new entries to the
     * agenda for completed arcs).
     */
    void addEntryToChart(Entry entry) {
	// 1. Add entry to chart
	int start = entry.getStart();
	int end = entry.getEnd();
	Debug.trace("Entering " + entry.getId() + ": " +
		    entry.getConstituent() + " \"" + tokensAsString(start, end) + "\"" +
		    " from " + start + " to " + end);
	chart.add(entry);
	// Check if entry is a complete parse (we could catch this at the
	// time the entry is added to the agenda, but then it wouldn't
	// be in the chart if we end up stopping the parse). A prioritized
	// agenda would look after this issue.
	if (entryIsCompleteParse(entry)) {
	    Debug.trace("  ** This is a complete parse");
	    nParsesFound += 1;
	}
	// 2. Add arcs for rules started by this entry
	fireRulesForEntry(entry);
	// 3. Extend any arcs for which this is the next constit
	extendArcsForEntry(entry);
    }
    /**
     * Return a string consisting of the input tokens of the sentence
     * being parsed within the given start-end range. This is used only
     * for the trace messages in addEntryToChart() (but is very informative
     * there).
     */
    String tokensAsString(int start, int end) {
	StringBuffer buf = new StringBuffer("");
	for (int i=start; i < end; i++) {
	    buf.append(tokens.get(i));
	    if (i < end-1) {
		buf.append(" ");
	    }
	}
	return buf.toString().toLowerCase();
    }
    /**
     * Returns true if the given Entry represents a complete parse of
     * the input. This method may be overidden by subclasses.
     */
    boolean entryIsCompleteParse(Entry entry) {
	return (entry.getEnd() == inputLength && entry.getStart() == 0 &&
		entry.getConstituent().getCategory().equals(SYMBOL_S));
    }
    /**
     * The Symbol ``S'', used for the syntactic category representing
     * complete parses.
     */
    final Symbol SYMBOL_S = Symbol.intern("S");
    /**
     * Step 2 of the bottom-up parsing algorithm (step 3 in second edition).
     * <P>
     * For newly-added constituent C from p1 to p2:
     * If C begins a rule R in the grammar, add an active arc for R
     * from p1 to p2.
     * That is, if R = ``X -> C X2 .. Xn'' is a rule in grammar, 
     * add an active arc ``X -> C * X2 ... XN''  from p1 to p2.
     */
    void fireRulesForEntry(Entry entry) {
	Constituent constit = entry.getConstituent();
	int start = entry.getStart();
	int end = entry.getEnd();
	Iterator rules = grammar.getRulesForConstituent(constit).iterator();
	try {
	    while (rules.hasNext()) {
		GrammarRule rule = (GrammarRule)rules.next();
		Arc arc = new Arc(rule, start, end, 1);
		addArcToChart(arc);
	    }
	} catch (NoSuchElementException ex) {
	}
    }
    /**
     * Step 3 of the bottom-up parsing algorithm (step 4, substep 2 in second
     * edition).
     * <P>
     * For newly-added constituent C from p1 to p2:
     * For any active arc A beginning at (some) p0 and ending at p1
     * (the beginning of C), if C is the next subconstituent of A,
     * then add an active arc from p0 to p2 with the updated rule from A.
     * That is, for any active arc of the form ``X -> X1 ... * C ... Xn''
     * from p0 to p1 (the beginning of C), add an active arc
     * ``X -> X1 ... C * ... Xn'' from p0 to p2.
     * <P>
     * Note: The description in the second edition of NLU phrases this
     * a bit differently, putting the ``completed arc'' condition test
     * here rather than handling the arcs from steps 2 and 3 uniformly.
     * The original description of the process works better in code.
     */
    void extendArcsForEntry(Entry entry) {
	Iterator arcs = chart.getArcsEndingAtPosition(entry.getStart()).iterator();
	try {
	    while (arcs.hasNext()) {
		Arc arc = (Arc)arcs.next();
		Arc newarc = extendArcWithEntry(arc, entry);
		if (newarc != null) {
		    addArcToChart(newarc);
		}
	    }
	} catch (NoSuchElementException ex) {
	}
    }
    /**
     * If the constituent of the given Entry matches the next Constituent
     * of the given Arc, then create and return a new Arc extending the given
     * one, otherwise return null.
     * This method may be overidden by subclasses.
     */
    Arc extendArcWithEntry(Arc arc, Entry entry) {
	Constituent constit = entry.getConstituent();
	Constituent nextconstit = arc.getNextConstituent();
	if (constit.getCategory().equals(nextconstit.getCategory())) {
	    return new Arc(arc.getRule(),
			   arc.getStart(), entry.getEnd(),
			   arc.getNext() + 1);
	} else {
	    return null;
	}
    }
    /**
     * Step 4 of the bottom-up parsing algorithm (step 4, substep 3 in second
     * edition).
     * <P>
     * If any of the active arcs added in steps 2 or 3 are completed rules,
     * add the new constituents, which are named by the left-hand side of
     * the rules, to the agenda.
     */
    void addArcToChart(Arc arc) {
	// Check for completed arc
	if (arc.isComplete()) {
	    // If so, add new entry for rhs to agenda
	    Entry entry = createEntryForCompletedArc(arc);
	    Debug.trace("  An " + entry.getConstituent().getCategory() + ", "
			+ entry.getId() +
			", from " + arc.getStart() + " to " + arc.getEnd() +
			" is created, by completing\n    " + arc);
	    agenda.add(entry);
	} else {
	    // Otherwise just add the arc
	    Debug.trace("  Adding arc " + arc);
	    chart.add(arc);
	}
    }
    /**
     * Create the new Entry to be added to the agenda for given
     * (completed) Arc. This method may be overidden by subclasses.
     */
    Entry createEntryForCompletedArc(Arc arc) {
	Constituent newconstit = arc.getRule().getLhs();
	Symbol id = createEntryId(newconstit.getCategory());
	return new Entry(id, newconstit, arc.getStart(), arc.getEnd());
    }
    /**
     * Test the basic bottom-up chart BottomUpChartParser.
     */
    public static void main(String argv[]) {
	main(argv, Lexicon.class, Grammar.class, BottomUpChartParser.class);
    }
    /**
     * For use by subclasses that need to specify different types
     * of lexicon, grammar, and parser.
     */
    public static void main(String argv[], Class lexiconClass, Class grammarClass, Class parserClass) {
	String USAGE = "usage: java [opts] " +
	    parserClass.getName() +
	    " [-trace true|false] [-debug true|false|io]" +
	    "[-l lexicon] [-g grammar]";
	try {
	    Lexicon lexicon = (Lexicon)lexiconClass.newInstance();
	    Grammar grammar = (Grammar)grammarClass.newInstance();
	    for (int i=0; i < argv.length; i++) {
		if (argv[i].equals("-trace")) {
		    i += 1;
		    if (argv[i].equals("true")) {
			Debug.setTracingEnabled(true);
		    } else if (argv[i].equals("false")) {
			Debug.setTracingEnabled(false);
		    } else {
			System.err.println(USAGE);
			System.exit(-1);
		    }
		} else if (argv[i].equals("-debug")) {
		    i += 1;
		    if (argv[i].equals("true")) {
			Debug.setDebug(true);
		    } else if (argv[i].equals("false")) {
			Debug.setDebug(false);
		    } else if (argv[i].equals("io")) {
			Debug.setDebugIO(true);
		    }
		} else if (argv[i].equals("-l")) {
		    lexicon.readFile(argv[++i]);
		} else if (argv[i].equals("-g")) {
		    grammar.readFile(argv[++i]);
		} else {
		    System.err.println(USAGE);
		    System.exit(-1);
		}
	    }
	    // Create parser
	    BottomUpChartParser parser = (BottomUpChartParser)parserClass.newInstance();
	    parser.setLexicon(lexicon);
	    parser.setGrammar(grammar);
	    // Read stdin and parse
	    BufferedReader stdin =
		new BufferedReader(new InputStreamReader(System.in));
	    while (true) {
		System.err.print("Input: ");
		String line = stdin.readLine();
		if (line == null) {
		    break;
		}
		parser.parse(line);
	    }
	} catch (EOFException ex) {
	} catch (Exception ex) {
	    ex.printStackTrace();
	}
    }
}
