/*
 * Grammar.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 25 Apr 2002
 * Time-stamp: <Thu May 22 15:25:10 EDT 2003 ferguson>
 */

import java.io.IOException;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * A grammar is a set of GrammarRules.
 * <P>
 * Implementation note: Since the main use of a Grammar is looking
 * up rules whose first constituent matches a newly-added entry, we
 * maintain the grammar rules indexed by category of their first
 * first rhs constituent.
 */
public class Grammar extends HashMap {
    /**
     * Returns a new (empty) Grammar.
     */
    public Grammar() {
    }
    /**
     * Adds the given GrammarRule to this Grammar. The Grammar is
     * indexed by syntactic catgeory of the first rhs constituent
     * to help in matching.
     */
    public void add(GrammarRule rule) {
	Constituent constit = (Constituent)rule.getRhs().get(0);
	Symbol category = constit.getCategory();
	// Get list of rules for this category
	List rules = (List)get(category);
	// Create list if necessary
	if (rules == null) {
	    rules = new LinkedList();
	    put(category, rules);
	}
	// Add rule to list of rules for this category
	rules.add(rule);
    }
    /**
     * Retrieve the List of GrammarRules whose first rhs constituent
     * matches the given Constituent.
     */
    public List getRulesForConstituent(Constituent constit) {
	Symbol category = constit.getCategory();
	List result = (List)get(category);
	if (result == null) {
	    result = new LinkedList();
	    put(category, result);
	}
	return result;
    }
    /**
     * Dump the contents of this Grammar to System.out.
     */
    public void dump() {
	dump(System.out);
    }
    /**
     * Dump the contents of this Grammar to the given PrintStream.
     */
    public void dump(PrintStream stream) {
	stream.println(description);
	Iterator keys = keySet().iterator();
	try {
	    while (keys.hasNext()) {
		Symbol category = (Symbol)keys.next();
		Iterator rules = ((List)get(category)).iterator();
		while (rules.hasNext()) {
		    GrammarRule rule = (GrammarRule)rules.next();
		    stream.println(rule.toString());
		}
	    }
	} catch (NoSuchElementException ex) {
	}
    }
    /**
     * The text description of this Grammar. This may be useful in
     * displays, tools, or just as a comment in the files.
     */
    String description;
    /**
     * Returns the text description of this Grammar.
     */
    public String getDescription() {
	return description;
    }
    /**
     * Sets the text description of this Grammar.
     */
    public void setDescription(String description) {
	this.description = description;
    }
    /**
     * Read the given filename and augment this Grammar.
     * @see GrammarFileHandler
     */
    public void readFile(String filename) throws IOException {
	new GrammarFileHandler(this).readFile(filename);
    }
}
