/*
 * HeadFeaturedGrammarFileHandler.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 10 May 2002
 * Time-stamp: <Fri May 23 11:22:22 EDT 2003 ferguson>
 *
 */

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * An HeadFeaturedGrammarFileHandler reads grammars whose constituents are
 * HeadFeaturedConstituents.
 * <P>
 * The BNF grammar we are trying to parse is something like the
 * following:
 * <PRE>
    Grammar <- START("Grammar") Description? HeadFeatures? GrammarRule* END("Grammar")
    Description <- START("Description") characters END("Description")
    HeadFeatures <- START("HeadFeatures") HeadFeature* END("HeadFeatures")
    HeadFeature <- START(CATEGORY) FEATNAME* END(CATEGORY)
    GrammarRule <- START("Rule") Constit END("Rule")
    Constit <- START(CATEGORY) Feature* END(CATEGORY)
    Feature <- START(FEATNAME) FeatureValue END(FEATNAME)
    FeatureValue <- characters
 * </PRE>
 * where CATEGORY is a syntactic category, and FEATNAME is a feature name.
 *<P>
 * The head constituent of a rule is marked with the HEAD attribute.
 */
public class HeadFeaturedGrammarFileHandler extends AugmentedGrammarFileHandler {
    /**
     * Create a new HeadFeaturedGrammarFileHandler to read into the given
     * HeadFeaturedGrammar.
     */
    public HeadFeaturedGrammarFileHandler(HeadFeaturedGrammar g) {
	super(g);
    }
    //
    // Finite-state machine parser starts here
    //
    static final int STATE_HEADFEATURES = 6;
    boolean isHead;
    public void startElement(String namespaceURI, String sName, String qName,
                             Attributes attrs)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("head: START " + ename + ", state=" + state);
	switch (state) {
	case STATE_GRAMMAR:
	    if (ename.equalsIgnoreCase("headfeatures")) {
		state = STATE_HEADFEATURES;
	    } else {
		super.startElement(namespaceURI, sName, qName, attrs);
	    }
	    return;
	case STATE_RULE:
	    // Head feature comes in as attribute (if given)
	    isHead = (XMLUtil.getAttributeValue(attrs, "head") != null);
	    break;
	case STATE_HEADFEATURES:
	    // Specifying head features for the category
	    cat = Symbol.intern(ename);
	    return;
	}
	// Unless handled above...
	super.startElement(namespaceURI, sName, qName, attrs);
    }
    public void endElement(String namespaceURI, String sName, String qName)
						throws SAXException {
        String ename = sName; // element name
        if (ename.equals("")) {
	    ename = qName; // namespaceAware = false
	}
	Debug.debugIO("head: END " + ename);
	switch (state) {
	case STATE_RULE:
	    // Add rule to grammar
	    HeadFeaturedConstituent lhs = (HeadFeaturedConstituent)constits.removeFirst();
	    List rhs = constits;
	    grammar.add(new HeadFeaturedGrammarRule(ruleId, lhs, rhs));
	    // Continue with grammar
	    state = STATE_GRAMMAR;
	    return;
	case STATE_CONSTIT:
	    // Make sure we're closing the constit we started
	    if (!ename.equalsIgnoreCase(cat.toString())) {
		ERROR("expected /" + cat + "; got /" + ename);
	    }
	    constits.add(new HeadFeaturedConstituent(cat, features, isHead));
	    // Consit is finished: continue with rule
	    state = STATE_RULE;
	    return;
	case STATE_HEADFEATURES:
	    if (ename.equalsIgnoreCase("HEADFEATURES")) {
		// End of specifying head features: rules should be next
		state = STATE_GRAMMAR;
	    } else {
		// Just end of one category's head features
	    }
	    return;
	}
	// Unless handled above...
	super.endElement(namespaceURI, sName, qName);
    }
    public void characters(char buf[], int offset, int len)
						throws SAXException
    {
	String text = new String(buf, offset, len).trim();
	if (text.equals("")) {
	    return;
	}
	Debug.debugIO("head: TEXT: \"" + text + "\"");
	switch (state) {
	case STATE_HEADFEATURES:
	    // List of feature names that are head features for cat
	    List names = new LinkedList();
	    StringTokenizer tokens = new StringTokenizer(text);
	    try {
		while (tokens.hasMoreTokens()) {
		    String token = tokens.nextToken();
		    names.add(Symbol.intern(token));
		}
	    } catch (NoSuchElementException ex) {
	    }
	    ((HeadFeaturedGrammar)grammar).setHeadFeatures(cat, names);
	    return;
	}
	// Unless handled above...
	super.characters(buf, offset, len);
    }

    public static void main(String argv[]) throws IOException {
	HeadFeaturedGrammar grammar = new HeadFeaturedGrammar();
	for (int i=0; i < argv.length; i++) {
	    grammar.readFile(argv[i]);
	}
	grammar.dump();
    }
}
