/*
 * Expression.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 24 Apr 2002
 * Time-stamp: <Wed May 14 12:20:23 EDT 2003 ferguson>
 */

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * Expressions are ordered sequences of Symbols, Variables, and
 * sub-Expressions (i.e., Unifiables).
 */
class Expression extends ArrayList implements List, Unifiable {
    /**
     * Adds the given Unifiable to this Expression (at the end).
     * We override our parent's method for this operation in order to
     * ensure that only Unifiables are added to Expressions. We will
     * throw a ClassCastException if the given object is not Unifiable.
     * This is better than having it come up later when we try to unify
     * the Expression.
     */
    public boolean add(Object x) {
	return super.add((Unifiable)x);
    }
    /**
     * Returns the printable representation of this Expression as a
     * String (using Lisp list syntax).
     */
    public String toString() {
	return Printing.listToString(iterator());
    }
    /**
     * Try to unify this Expression with another Unifiable.
     * Two Expressions unify if their respective components unify.
     */
    public boolean unifyWith(Unifiable other, BindingSet bindings) {
	if (other instanceof Variable) {
	    // Expression might unify with Variable: go see
	    return other.unifyWith(this, bindings);
	} else if (!(other instanceof Expression)) {
	    // Otherwise Expression unifies only with another Expression
	    return false;
	}
	// Ok, we've got two Expressions
	Expression othere = (Expression)other;
	// Must be of the same length
	if (size() != othere.size()) {
	    return false;
	}
	// And respective items must unify (with consistent bindings)
	Iterator myitems = iterator();
	Iterator otheritems = othere.iterator();
	try {
	    while (myitems.hasNext()) {
		Unifiable myelement = (Unifiable)myitems.next();
		Unifiable otherelement = (Unifiable)otheritems.next();
		if (!myelement.unifyWith(otherelement, bindings)) {
		    return false;
		}
	    }
	} catch (NoSuchElementException ex) {
	}
	return true;
    }
    /**
     * Returns a new Expression which is a copy of this one with the
     * given bindings applied. This just means traversing the elements
     * of the Expression and recursively calling copyWithBindings().
     */
    public Unifiable copyWithBindings(BindingSet bindings) {
	Expression result = new Expression();
	Iterator elts = iterator();
	try {
	    while (elts.hasNext()) {
		Unifiable elt = (Unifiable)elts.next();
		result.add(elt.copyWithBindings(bindings));
	    }
	} catch (NoSuchElementException ex) {
	}
	return result;
    }
}
