/*
 * AssociationList.java
 *
 * George Ferguson, ferguson@cs.rochester.edu, 15 Apr 2002
 * Time-stamp: <Wed May 14 12:04:33 EDT 2003 ferguson>
 */

import java.util.AbstractMap;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;

/**
 * An AssociationList (or ``alist'') represents a set of values indexed
 * by keys, and so implements the Map interface. The difference is that
 * AssociationList is implemented using a LinkedList of the key-value
 * pairs. One could use a HashMap, but the overhead would probably be
 * higher for small alists.
 */
public class AssociationList extends AbstractMap implements Map {
    /**
     * LinkedList to hold entries in AssociationList.
     */
    LinkedList entries;
    /**
     * Creates a new empty AssociationList.
     * This constructor is required by the conventions of the Collection
     * interface.
     */
    public AssociationList() {
	entries = new LinkedList();
    }
    /**
     * Creates a new AssociationList with the same contents as the given
     * Collection.
     * This constructor is required by the conventions of the Collection
     * interface.
     */
    public AssociationList(Collection c) {
	entries = new LinkedList(c);
    }
    /**
     * Returns a set view of the mappings contained in this map.
     * <BR>
     * This implementation returns a LinkedListSet whose elements
     * consist of OrderedPair's.
     */
    public Set entrySet() {
	Set s = new LinkedListSet();
	Iterator iter = entries.iterator();
	try {
	    while (iter.hasNext()) {
		s.add(iter.next());
	    }
	} catch (NoSuchElementException ex) {
	}
	return s;
    }
    /**
     * Associates the specified value with the specified key in this map.
     * <BR>
     * This is equivalent to the Lisp form:
     *   (setq LIST (cons (cons KEY VALUE) LIST))
     * or the shorthand using ACONS.
     * <BR>
     * Implementation note: Technically (in Lisp), an alist does not need
     * to check for the prior presence of an entry for a given key, since
     * the way items are looked up (by ASSOC) guarantees that only the most
     * recently added entry will be returned. However, since we are
     * implementing the full Map interface, we need to be more careful
     * to avoid problems with access through iterators and the like (note
     * that this would also be a problem if one used MAPCAR and the like
     * on an alist).
     * <BR>
     * Another technical implementation note is that by using MapEntry.setValue()
     * on the existing entry (if there is one), we modify the entry in a
     * way that might be visible to other code (like using (SETF CDR) rather
     * than ACONS in Lisp).
     */
    public Object put(Object key, Object value) {
	Iterator items = entries.iterator();
	try {
	    while (items.hasNext()) {
		Map.Entry e = (Map.Entry)items.next();
		if (e.getKey().equals(key)) {
		    e.setValue(value);
		    // What to return?
		    return value;
		}
	    }
	} catch (NoSuchElementException ex) {
	}
	entries.add(new OrderedPair(key, value));
	return value;
    }
}
